import time
from typing import List, Optional, Callable
import numpy as np
import torch
from tqdm import tqdm
from sklearn.decomposition import PCA
import math
from scipy.spatial import distance
import faiss
# import faiss_gpu



EMBEDDING_TYPE = np.ndarray
defualt_dist_func = lambda u, v: np.sqrt(np.power(u - v, 2).sum(-1))  # Use mse distance as default
dist_func_sqrt = lambda u, v: np.sqrt(np.abs(u - v)).sum(-1)  # Use mse distance as default
defualt_dist_func = dist_func_sqrt # todo tmp check
defualt_dist_func_element_wise = lambda u, v: u - v  # Use  distance as default
# defualt_dist_func_one_to_many = lambda u, v: np.mean(defualt_dist_func(np.tile(u, (v.shape[0], 1)), v))
defualt_dist_func_one_to_many = lambda u, v: defualt_dist_func(np.tile(u, (v.shape[0], 1)), v)


def calc_dist_matrix(x: EMBEDDING_TYPE, y: EMBEDDING_TYPE,
                     dist_func: Optional[Callable[[EMBEDDING_TYPE, EMBEDDING_TYPE], EMBEDDING_TYPE]] or str = None,
                     low_ram: bool = True) -> EMBEDDING_TYPE:
    """
    A function to get the distance between embedding matrices

    :param x: Embedding in the form of [n ,t] matrix, where n is the sample dimension and t is the features dimension
    :param y: Embedding in the form of [m ,t] matrix, where m is the sample dimension and t is the features dimension
    :param dist_func: Function that gets as input x, y type input return their element wise distance
    :param low_ram: if set to True will calculate the distances one row at the time. Default is mse distance
    :return: a distance matrix in the shape of [n, m] where element dist[j,s] is the distance of x[j] to y[s]
    """
    if dist_func is None:
        dist_func = 'euclidean'  # default

    # curr_time = time.time()
    dist_matrix = distance.cdist(x, y, dist_func)
    # res_time = time.time() - curr_time
    # print(f'{res_time = }')
    # curr_time = time.time()
    # dist_matrix = distance.cdist(x, y, 'euclidean')
    # res_time = time.time() - curr_time
    # print(f'{res_time = }')

    # n = x.shape[0]
    # m = y.shape[0]
    # x = np.tile(np.expand_dims(x, axis=1), (1, m, 1))
    # y = np.tile(np.expand_dims(y, axis=0), (n, 1, 1))
    # curr_time = time.time()

    # low_ram = False
    # if not low_ram:
    #     dist_matrix = dist_func(x, y)
    # else:
    #     dist_matrix = []
    #     for p_ind in tqdm(range(x.shape[0]), desc='low_ram is set to True. Trying one row at the time'):
    #         dist_matrix += [np.expand_dims(dist_func(x[p_ind], y[p_ind]), 0)]
    #     dist_matrix = np.concatenate(dist_matrix, 0)
    #
    # res_time = time.time() - curr_time
    # print(f'old {res_time = }')
    return dist_matrix


def get_k_neighbours_indexes_for_each_sample(dist_mat: EMBEDDING_TYPE, top_k: Optional[int] = 5,
                                             do_neighbours_one_by_one: bool = False) -> EMBEDDING_TYPE:
    """
    a function to get the indecis of the top k elements in each row of dist mat
    :param dist_mat: EMBEDDING_TYPE object of size [n, m]
                     where n is the number of source embedding to calculate the distance from,
                     and m is the number of embedding destination to calculate the the distance to.
    :param top_k: Optional number of neighbours for each sample to return, default is 5
    :return: an ndarray of size [n,top_k+1] where foreach sample in n,
     the indexes of the top k closest neighbours from m are returned.
     the first element will always be the current sample index,
    """
    if not do_neighbours_one_by_one:
        top_k_indices = np.argsort(dist_mat, axis=1)[:, :top_k + 1]  # get the top k indices for each row in dist mat
    else:
        one_closest_neighbour_s = np.argsort(dist_mat, axis=1)[:, 1]  # get the top k indices for each row in dist mat
        top_k_indices = np.argsort(dist_mat, axis=1)[:, :top_k + 2]  # get the top k indices for each row in dist mat
        # top_k_dists = [dist_mat_row[k_indices] for dist_mat_row, k_indices in zip(dist_mat, top_k_indices)]  #todo try to implemet grahp to added the closse among a the tree of negiburs not just the last one

        top_k_indices_final = [[cur] for cur in range(len(one_closest_neighbour_s))]
        for ind in range(top_k + 1):
            for sample_index, already_included_neighbours in enumerate(top_k_indices_final):
                neighbour_to_check_index = already_included_neighbours[-1]
                neighbour_to_choose_from = top_k_indices[neighbour_to_check_index]
                already_use_neighbour_bool = [neighbour in already_included_neighbours for neighbour in neighbour_to_choose_from]
                closest_neighbour_not_already_chosen_index = next((i for i, j in enumerate(already_use_neighbour_bool) if not j), None)
                closet_neighbour_to_not_already_chosen = neighbour_to_choose_from[closest_neighbour_not_already_chosen_index]
                top_k_indices_final[sample_index] += [closet_neighbour_to_not_already_chosen]
        for c_ind, cur_top_k in enumerate(top_k_indices_final):
            top_k_indices_final[c_ind] = np.expand_dims(np.array(cur_top_k), axis=0)
        top_k_indices_final = np.concatenate(top_k_indices_final, axis=0)
        top_k_indices = top_k_indices_final
    return top_k_indices
def get_k_neighbours_indexes_for_each_sample(dist_mat: EMBEDDING_TYPE, top_k: Optional[int] = 5,
                                             do_neighbours_one_by_one: bool = False) -> EMBEDDING_TYPE:
    """
    a function to get the indecis of the top k elements in each row of dist mat
    :param dist_mat: EMBEDDING_TYPE object of size [n, m]
                     where n is the number of source embedding to calculate the distance from,
                     and m is the number of embedding destination to calculate the the distance to.
    :param top_k: Optional number of neighbours for each sample to return, default is 5
    :return: an ndarray of size [n,top_k+1] where foreach sample in n,
     the indexes of the top k closest neighbours from m are returned.
     the first element will always be the current sample index,
    """
    if not do_neighbours_one_by_one:
        top_k_indices = np.argsort(dist_mat, axis=1)[:, :top_k + 1]  # get the top k indices for each row in dist mat
    else:
        one_closest_neighbour_s = np.argsort(dist_mat, axis=1)[:, 1]  # get the top k indices for each row in dist mat
        top_k_indices = np.argsort(dist_mat, axis=1)[:, :top_k + 2]  # get the top k indices for each row in dist mat
        # top_k_dists = [dist_mat_row[k_indices] for dist_mat_row, k_indices in zip(dist_mat, top_k_indices)]  #todo try to implemet grahp to added the closse among a the tree of negiburs not just the last one

        top_k_indices_final = [[cur] for cur in range(len(one_closest_neighbour_s))]
        for ind in range(top_k + 1):
            for sample_index, already_included_neighbours in enumerate(top_k_indices_final):
                neighbour_to_check_index = already_included_neighbours[-1]
                neighbour_to_choose_from = top_k_indices[neighbour_to_check_index]
                already_use_neighbour_bool = [neighbour in already_included_neighbours for neighbour in neighbour_to_choose_from]
                closest_neighbour_not_already_chosen_index = next((i for i, j in enumerate(already_use_neighbour_bool) if not j), None)
                closet_neighbour_to_not_already_chosen = neighbour_to_choose_from[closest_neighbour_not_already_chosen_index]
                top_k_indices_final[sample_index] += [closet_neighbour_to_not_already_chosen]
        for c_ind, cur_top_k in enumerate(top_k_indices_final):
            top_k_indices_final[c_ind] = np.expand_dims(np.array(cur_top_k), axis=0)
        top_k_indices_final = np.concatenate(top_k_indices_final, axis=0)
        top_k_indices = top_k_indices_final
    return top_k_indices

def get_partial_eig_on_embedding(data: EMBEDDING_TYPE, partial_size: Optional[int] = 2):
    """
    :param data: Matrix of size[m ,n] where n is the features dimensions and m the sample dimensions
    :param partial_size: Size to each features group to divide the futures to
    :return: sub_V, sub_S the eig values and vectors for each partial feature group.
            sub_V: the partial eig vectors size of [u,v,w],
                                                    u -> the index of the partial eig vec,
                                                    v -> the eig vec elements
                                                    w -> more eig vecs for the same index.
            sub_S : the partial eig valus size of [u, w], u -> the index of the partial eig vec,
                                                    w -> more eig vecs for the same index.
            example: data = [11, 500]
                     sub_V, sub_S = get_partial_eig_vecs(data=data, partial_size=5)
                     sub_V = [100 5, 5]
                     sub_S = [100, 5]
    """
    start_ind = 0
    sub_V_s = []
    sub_S_s = []
    pca = PCA(n_components=partial_size)
    while start_ind < data.shape[1] - 1:
        end_ind = start_ind + partial_size
        end_ind = end_ind if end_ind < data.shape[1] else (data.shape[1])
        vec_len = end_ind - start_ind
        data_for_eig = data[:, start_ind:end_ind]
        # if the embedding vector is not divisible by partial size we need to add some zeros for the pce function.
        if vec_len < partial_size:
            add_zeros = np.zeros((data.shape[0], partial_size - vec_len))
            data_for_eig = np.concatenate((data_for_eig, add_zeros), axis=1)
        pca.fit(data_for_eig)
        cur_V = pca.components_  # row are samples
        # cur_V = pca.components_.T
        cur_S = pca.singular_values_
        start_ind = end_ind
        sub_V_s += [np.expand_dims(cur_V, axis=0)]  # V columns are eig vec, rows are different features
        sub_S_s += [np.expand_dims(cur_S, axis=0)]

    if len(sub_V_s) == data.shape[1] - 1 and not sub_V_s[0].shape[1] == end_ind:  # the eig size is 1 and we need to add the last one
        end_ind = start_ind + partial_size
        pca.fit(data[:, start_ind:end_ind])
        cur_V = pca.components_.T
        cur_S = pca.singular_values_
        sub_V_s += [cur_V.unsqueeze(0)]  # V columns are eig vec, rows are different features
        sub_S_s += [cur_S.unsqueeze(0)]
    # add zeros to the eig vec
    sub_V_s[-1] = np.concatenate(
        (sub_V_s[-1], np.zeros((1, sub_V_s[0].shape[1] - sub_V_s[-1].shape[1], sub_V_s[-1].shape[2]))), axis=1)
    sub_V_s[-1] = np.concatenate(
        (sub_V_s[-1], np.zeros((1, sub_V_s[-1].shape[1], sub_V_s[0].shape[2] - sub_V_s[-1].shape[2]))), axis=2)
    sub_S_s[-1] = np.concatenate((sub_S_s[-1], np.zeros((1, sub_S_s[0].shape[1] - sub_S_s[-1].shape[1]))), axis=1)
    sub_V_s = np.concatenate(sub_V_s, axis=0)
    sub_S_s = np.concatenate(sub_S_s, axis=0)
    return sub_V_s, sub_S_s


def get_neighbours_partial_eig_by_top_k(embeddings: EMBEDDING_TYPE, top_k_mat: EMBEDDING_TYPE,
                                        partial_size: int = 5) -> [EMBEDDING_TYPE, EMBEDDING_TYPE]:



    # num_of_elem_in_eig_vec = top_k_mat.shape[1]
    num_of_partial_vec = math.ceil(embeddings.shape[1] / partial_size)
    data_eig_vec_partial_n = np.zeros((top_k_mat.shape[0], num_of_partial_vec, partial_size, partial_size))
    data_eig_val_partial_n = np.zeros((top_k_mat.shape[0], num_of_partial_vec, partial_size))

    for ind in tqdm(range(top_k_mat.shape[0])):
        curr_data = embeddings[top_k_mat[ind], :]
        sub_v_s, sub_s_s = get_partial_eig_on_embedding(data=curr_data, partial_size=partial_size)
        data_eig_vec_partial_n[ind, :, :, :] = sub_v_s
        data_eig_val_partial_n[ind, :, :] = sub_s_s
    return data_eig_vec_partial_n, data_eig_val_partial_n


def normalize_eig_vec_by_eig_val_reshape(eig_vec_s: EMBEDDING_TYPE, eig_val_s: EMBEDDING_TYPE,
                                         embedding_size: int, epsilon: float = 1e-4) -> EMBEDDING_TYPE:
    """

    :param eig_vec_s: matrix in the shape of [n,p,v,f] where n is the number nearest samples, p is the partial eig vec index, v is the number of eig vector and f the nuber of features in each partial eig vec
    :param eig_val_s: matrix of shape [n,p,v] the eigen values corresponding to the eigen vectors
    :param embedding_size: the overall size of the feature in each samples embedding
    :param epsilon: a small number to use in the calculation so that the return values will not reach inf
    :return:
    Ebedding TYPE matrix of the concatenated partial eigen vector reweighed by their eigen values
    """
    eig_val_s_for_normalization = np.tile(np.expand_dims(eig_val_s, axis=3), (1, 1, 1, eig_vec_s.shape[3]))
    eig_vec_normalize_by_val = eig_vec_s / (eig_val_s_for_normalization + epsilon)
    # eig_vec_normalize_by_val_reshaped = eig_vec_normalize_by_val.reshape(
    #     (eig_vec_normalize_by_val.shape[0], -1, eig_vec_normalize_by_val.shape[2]))

    eig_vec_normalize_by_val_reshaped = eig_vec_normalize_by_val.reshape(
        (eig_vec_normalize_by_val.shape[0], eig_vec_normalize_by_val.shape[2], -1))
    eig_vec_normalize_by_val_reshaped = np.moveaxis(eig_vec_normalize_by_val_reshaped, 1, -1)
    eig_vec_normalize_by_val_reshaped = eig_vec_normalize_by_val_reshaped[:, :embedding_size, :]  # if the ebedding size is not a devisibal by eig size then some zeros are added
    return eig_vec_normalize_by_val_reshaped


def sample_score_using_normalized_eig_mat_and_dist_mat(element_dist: EMBEDDING_TYPE,
                                                       normalized_eig_mat: EMBEDDING_TYPE) -> float:
    element_wise_diff_mat_repeat = np.tile(np.expand_dims(element_dist, axis=2), (1, 1, normalized_eig_mat.shape[2]))

    partial_eig_dist_normed = element_wise_diff_mat_repeat * normalized_eig_mat

    # the direction of the eig vs the diff can result in negative or positive
    eig_vec_size = normalized_eig_mat.shape[2]
    number_of_partial_features = normalized_eig_mat.shape[1] // normalized_eig_mat.shape[2]
    number_of_partial_features = number_of_partial_features if normalized_eig_mat.shape[1] % normalized_eig_mat.shape[2] == 0 else number_of_partial_features + 1  #  if the number of featuer in each partial vec is not devisibal by the overall number of featuers
    partial_eig_dist_normed_res = np.zeros((normalized_eig_mat.shape[0], number_of_partial_features, normalized_eig_mat.shape[2]))
    for res_ind in range(partial_eig_dist_normed_res.shape[1]):
        partial_ind = res_ind * eig_vec_size
        partial_end_ind = partial_ind + eig_vec_size if partial_ind + eig_vec_size <= partial_eig_dist_normed.shape[1] else partial_eig_dist_normed.shape[1]
        # print(f'{partial_eig_dist_normed.shape = }')
        partial_eig_dist_normed_res[:, res_ind, :] = np.abs((partial_eig_dist_normed[:, partial_ind: partial_end_ind, :].mean(1)))
    res_to_each_sample = partial_eig_dist_normed_res.mean((1, 2))
    return res_to_each_sample


def normalized_dist_by_partial_eigs(embedding_test: EMBEDDING_TYPE, embedding_train: EMBEDDING_TYPE,
                                    neighbours_indices_mat: EMBEDDING_TYPE,
                                    neighbours_of_neighbours_indices_mat: EMBEDDING_TYPE,
                                    partial_size: Optional[int] = None,
                                    partial_eig_vec_s: Optional[EMBEDDING_TYPE] = None,
                                    partial_eig_val_s: Optional[EMBEDDING_TYPE] = None,
                                    dist_func_element_wise: Optional[
                                        Callable[[EMBEDDING_TYPE, EMBEDDING_TYPE], EMBEDDING_TYPE]] = None,
                                    do_local: bool = False) -> List[float]:
    assert (partial_size is not None) or ((partial_eig_vec_s is not None) and (partial_eig_val_s is not None)), "Need to defind partial_eig_ or partial_size"
    partial_size = partial_size if partial_size != -1 else embedding_test.shape[1]
    if partial_eig_vec_s is not None:
        partial_size = partial_eig_vec_s.shape[2]
    if dist_func_element_wise is None:
        dist_func_element_wise = defualt_dist_func_element_wise

    test_embedding_for_dist = np.tile(np.expand_dims(embedding_test, axis=1), (1, neighbours_indices_mat.shape[1], 1))
    all_res = []
    for curr_ind, cur_neighbours_indices in tqdm(enumerate(neighbours_indices_mat), desc='Normalizing samples', total=neighbours_indices_mat.shape[0]):
        cur_neighbours_embedding = embedding_train[cur_neighbours_indices, :]
        curr_test_embedding = test_embedding_for_dist[curr_ind, :, :]
        if not do_local:
            cur_eig_vec_s = partial_eig_vec_s[cur_neighbours_indices, :, :, :]
            cur_eig_val_s = partial_eig_val_s[cur_neighbours_indices, :, :]
        else:
            partial_vec_local_curr, partial_val_local_curr = get_partial_eig_on_embedding(cur_neighbours_embedding, partial_size=partial_size)
            cur_eig_vec_s = np.tile(np.expand_dims(partial_vec_local_curr, 0), (neighbours_indices_mat.shape[1], 1, 1, 1))
            cur_eig_val_s = np.tile(np.expand_dims(partial_val_local_curr, 0), (neighbours_indices_mat.shape[1], 1, 1))

        cur_eig_vec_normalize_by_val_reshaped = normalize_eig_vec_by_eig_val_reshape(eig_vec_s=cur_eig_vec_s,
                                                                                     eig_val_s=cur_eig_val_s,
                                                                                     embedding_size=
                                                                                     embedding_train.shape[1])
        curr_element_wise_diff_mat = dist_func_element_wise(curr_test_embedding, cur_neighbours_embedding)
        res_to_sample = sample_score_using_normalized_eig_mat_and_dist_mat(element_dist=curr_element_wise_diff_mat,
                                                                           normalized_eig_mat=cur_eig_vec_normalize_by_val_reshaped)
        curr_anomaly_score = res_to_sample.mean()

        all_res += [curr_anomaly_score]

    return all_res


def get_new_embedding_order(embedding, set_size=-1):
    new_order = list(range(embedding.shape[1]))
    prev_vec_to_comp = 0
    do_second_order = True
    corr_mat_data = np.abs(np.corrcoef(embedding.T))
    corr_mat_data = np.nan_to_num(corr_mat_data)
    # print('test num to num')
    corr_mat_data_changing = np.copy(corr_mat_data)
    corr_mat_data_changing -= 2 * np.eye(corr_mat_data_changing.shape[0])
    corr_mat_data_changing = np.append(corr_mat_data_changing, np.expand_dims(
        np.arange(start=0, stop=corr_mat_data_changing.shape[0], step=1), 0), axis=0)
    for cur_corr_ind in range(embedding.shape[1]):
        elem_id = new_order[cur_corr_ind]
        # print('='*50)
        # print(f'{cur_corr_ind = }')
        # print(f'1: {new_order[:20] = }')
        to_max_vec = corr_mat_data_changing[:-1, :][corr_mat_data_changing[-1, :] == elem_id]
        if do_second_order:
            to_max_vec += prev_vec_to_comp
        if set_size > -1:
            if (cur_corr_ind + 1) % set_size == 0 and cur_corr_ind > 0:  # start a new set
                index_to_ignore = np.argmin(to_max_vec - prev_vec_to_comp)
                to_max_vec[0, index_to_ignore] = to_max_vec.max() + 1
                best_next_ind = np.argmin(to_max_vec)  # find the elem id by last index because the indexes are chaining
            else:
                best_next_ind = np.argmax(to_max_vec)  # find the elem id by last index because the indexes are chaining
        else:
            best_next_ind = np.argmax(to_max_vec)  # find the elem id by last index because the indexes are chaining


        best_next_elem = int(corr_mat_data_changing[-1, best_next_ind])
        # print(f': {best_next_ind = }')
        # print(f': {best_next_elem = }')
        new_order.remove(best_next_elem)
        new_order.insert(cur_corr_ind + 1, best_next_elem)
        # print(f'2: {new_order[:20] = }')

        remove_ind = np.argmax(corr_mat_data_changing[-1, :] == elem_id)
        corr_mat_data_changing = np.delete(corr_mat_data_changing, remove_ind, axis=1)
        if set_size > -1:
            if (cur_corr_ind+1) % set_size == 0 and cur_corr_ind > 0:  # start a new set
                prev_vec_to_comp = 0
                # print('-'*50)
            else:
                prev_vec_to_comp = np.copy(corr_mat_data_changing[remove_ind, :])
        else:
            prev_vec_to_comp = np.copy(corr_mat_data_changing[remove_ind, :])
        corr_mat_data_changing = np.delete(corr_mat_data_changing, remove_ind, axis=0)
    new_data_reorder = np.copy(embedding)
    new_data_reorder = new_data_reorder[:, new_order]
    return {'new_embedding': new_data_reorder, 'new_order': new_order}

class Normalized_KNN:
    """
    Get Normalized KNN distances between two embeddings based on different normalization schemes
    """
    def __init__(self,
                 number_of_neighbours: int = 10,
                 number_of_neighbours_of_neighbours: int = 10,
                 partial_size: int = 8,
                 embedding: Optional[EMBEDDING_TYPE] = None,
                 first_global_normalization: bool = True,
                 do_neighbours_one_by_one: bool = True,
                 do_reorder: bool = True,
                 do_weighted_eig: bool = True,
                 do_knnn_plus_knn=False,
                 post_dist_func: str = None,
                 eigen_power=-0.5,
                 ):
        self.partial_size: int = partial_size
        self.number_of_neighbours: int = number_of_neighbours
        self.number_of_neighbours_of_neighbours: int = number_of_neighbours_of_neighbours
        self.first_global_normalization = first_global_normalization
        self.do_neighbours_one_by_one = do_neighbours_one_by_one
        self.do_reorder: bool = do_reorder
        self.do_weighted_eig: bool = do_weighted_eig
        self.do_knnn_plus_knn: bool = do_knnn_plus_knn
        self.do_faiss_reorder = False
        self.eigen_power = eigen_power # 0.5 # 2 # -0.5
        if post_dist_func is not None and 'sqrt' in post_dist_func:
            self.post_dist_func = lambda x: np.sqrt(x)
        else:
            self.post_dist_func = lambda x: x

        if embedding is not None:
            self.fit(embedding=embedding)

    def fit(self, embedding: EMBEDDING_TYPE) -> None:
        """
        Calculate the necessary statistics for the embedding normalization
        :param embedding: mat of shape [n, m] where n is the sample dimension and m the features dimension
        :return: None
        """
        self.embedding: EMBEDDING_TYPE = embedding
        assert self.partial_size <= self.embedding.shape[1], 'partial size should be bigger than the embedding space'
        self.num_of_samples = self.embedding.shape[0]
        self.partial_size = self.partial_size if self.partial_size != -1 else self.embedding.shape[1]
        self.dim = self.embedding.shape[1]
        if self.do_reorder:
            if not self.do_faiss_reorder:
                reorder_res = get_new_embedding_order(embedding=self.embedding, set_size=self.partial_size)  # set_size=-1) # self.partial_size)
                self.embedding = reorder_res['new_embedding']
                self.new_order = reorder_res['new_order']

                # corr_mat_data = np.nan_to_num(np.abs(np.corrcoef(self.embedding.T)))
                # import plotly.express as px
                # fig = px.imshow(corr_mat_data)
                # fig.show()

            else:
                self.opq_mat = faiss.OPQMatrix(d=self.embedding.shape[-1], M=self.partial_size)
                self.opq_mat.train(np.ascontiguousarray(self.embedding))
                self.embedding = self.opq_mat.apply_py(np.ascontiguousarray(self.embedding))

            # print('end')



        self.embedding = np.ascontiguousarray(self.embedding)

        # self.partial_vec_global_s, self.partial_val_global_s = get_partial_eig_on_embedding(data=self.embedding,
        #                                                                                     partial_size=min(partial_size, num_of_samples))
        # todo add back and change acourding to faiss
        # if self.first_global_normalization:  # change the dist mat according to the normalized embedding
        #
        #     global_eig_vec_normalize_by_val_reshaped = normalize_eig_vec_by_eig_val_reshape(
        #         eig_vec_s=np.expand_dims(self.partial_vec_global_s, 0),
        #         eig_val_s=np.expand_dims(self.partial_val_global_s, 0),
        #         embedding_size=self.embedding.shape[1])
        #     global_eig_vec_normalize_by_val_reshaped = np.tile(global_eig_vec_normalize_by_val_reshaped,
        #                                                        (self.embedding.shape[0], 1, 1))
        #     self.dist_mat = calc_dist_matrix(
        #         x=self.embedding * global_eig_vec_normalize_by_val_reshaped.sum(2),
        #         y=self.embedding * global_eig_vec_normalize_by_val_reshaped.sum(2))
        #     # self.embedding *= global_eig_vec_normalize_by_val_reshaped.sum(2)
        # else:
        #     self.dist_mat = calc_dist_matrix(x=self.embedding, y=self.embedding)

        # self.top_k_indexes_train = get_k_neighbours_indexes_for_each_sample(dist_mat=self.dist_mat,
        #                                                                     top_k=self.number_of_neighbours_of_neighbours,
        #                                                                     do_neighbours_one_by_one=self.do_neighbours_one_by_one)

        #on faiss
        #on faiss
#         index = faiss.GpuIndexFlatL2(self.embedding.shape[1]) # todo check

        self.embedding = np.ascontiguousarray(self.embedding)
        self.embedding = np.float32(self.embedding)
        self.embedding_original = self.embedding.copy()

        if self.first_global_normalization:  # change the dist mat according to the normalized embedding # todo add global normalization with partial size
            self.whiten = faiss.PCAMatrix(d_in=self.dim, d_out=self.dim, eigen_power=self.eigen_power) # after transformation the components are multiplied by eigenvalues^eigen_power =0: no whitening =-0.5: full whitening
            self.whiten.epsilon = 1e-5
            # self.whiten = faiss.PCAMatrix(d_in=self.dim, d_out=self.dim, eigen_power=-2) # after transformation the components are multiplied by eigenvalues^eigen_power =0: no whitening =-0.5: full whitening
            self.whiten.train(self.embedding)
            # global_eig_vals = faiss.vector_to_array(self.whiten.eigenvalues)
            # print(f'{global_eig_vals = }')

            self.embedding_new = self.whiten.apply_py(self.embedding) # TODO add transform to test as well
            # self.embedding_new = np.nan_to_num(self.embedding_new) # tODO maybe check min berfor nan to num or add epsilin in fit
            if np.isnan(self.embedding_new).any():
                print(f'have NONE: {np.isnan(self.embedding_new).any()}')
                self.embedding_new = np.nan_to_num(self.embedding_new)
            self.embedding = self.embedding_new

            # tmp

            # import matplotlib.pyplot as plt
            # corr_mat_data = np.abs(np.corrcoef(self.embedding.T))
            # corr_mat_data = np.nan_to_num(corr_mat_data)
            # plt.imshow(corr_mat_data)
            # plt.show()


        self.cpu_index = faiss.IndexFlatL2(self.dim)
        # self.gpu_index = faiss.index_cpu_to_all_gpus(index=self.cpu_index)
        self.index = self.cpu_index # self.gpu_index # TODO !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! remove to use gpu
        # sometime the indexes needs to be trained like clustering this is how to check:
        # index.is_trained
        self.index.add(np.ascontiguousarray(self.embedding))

        # check size of the embedingns
        print(f'{self.index.ntotal = }')
        import time
        start = time.time()
        D, I = self.index.search(np.ascontiguousarray(self.embedding), self.number_of_neighbours_of_neighbours)
        end = time.time()
        print(f'search nonon indies time: {end - start}') # 706 sec for IndexFlatL2
        #    index = faiss.IndexFlatL2
        self.top_k_indexes_train = I

        # neighbours_embeddings = self.embedding[I, :]
        # neighbours_embeddings_quantized = [np.expand_dims(neighbours_embeddings[:, :, start_ind: start_ind+partial_size],2) for start_ind in range(0, neighbours_embeddings.shape[-1], partial_size)]
        # neighbours_embeddings_quantized = np.concatenate(neighbours_embeddings_quantized, axis=2)
        # neighbours_embeddings_quantized_whitening = []
        # for cur_neighbours in tqdm(neighbours_embeddings_quantized):
        #     neighbours_embeddings_quantized_whitening_quant = []
        #     for qunt_ind in range(neighbours_embeddings_quantized.shape[2]):
        #         cur_d_in = cur_neighbours.shape[-1]
        #         cur_d_out = cur_d_in
        #         curr_whitening = faiss.PCAMatrix(d_in=cur_d_in, d_out=cur_d_out, eigen_power=-0.5) # after transformation the components are multiplied by eigenvalues^eigen_power =0: no whitening =-0.5: full whitening
        #         # curr_whitening.train(cur_neighbours)
        #         curr_whitening.train(np.ascontiguousarray(cur_neighbours[:,qunt_ind,:]))
        #         neighbours_embeddings_quantized_whitening_quant += [curr_whitening]
        #     neighbours_embeddings_quantized_whitening += [neighbours_embeddings_quantized_whitening_quant]



        #on faiss
        #on faiss



        # if self.number_of_neighbours_of_neighbours >= partial_size:
        #     partial_vec_s, partial_val_s = get_neighbours_partial_eig_by_top_k(embeddings=self.embedding,
        #                                                                        top_k_mat=self.top_k_indexes_train,
        #                                                                        partial_size=partial_size)
        # else:
        #     partial_vec_s, partial_val_s = None, None
        #
        # self.partial_vec_s = partial_vec_s
        # self.partial_val_s = partial_val_s

    def __call__(self, test_embedding: EMBEDDING_TYPE,
                 do_global: bool = True, do_local: bool = True, do_knnn: bool = True, do_knn: bool = True) -> List[float]:
        """
        Get the scores on the test_embedding using normalized knn, relative to the saved embedding
        :param test_embedding: mat of shape [n, m] where n is the sample dimension and m the features dimension.
        :return: n size List where each element is the scores of a test embedding sample
        """
        assert self.embedding is not None, "Call .fit() before calculation the results"
        return_dict = {}
        if not self.do_faiss_reorder:
            test_embedding = test_embedding[:, self.new_order] if self.do_reorder else test_embedding
            test_embedding = np.ascontiguousarray(test_embedding)
        else:
            test_embedding = self.opq_mat.apply_py(np.ascontiguousarray(test_embedding))

        test_embedding = np.float32(test_embedding)
        if do_knn:
            knn_index_cpu = faiss.IndexFlatL2(self.dim)
            # knn_gpu_index = faiss.index_cpu_to_all_gpus(index=self.cpu_index)
            knn_index = knn_index_cpu
            knn_index.add(np.ascontiguousarray(self.embedding_original))
            D_test_knn, _ = knn_index.search(np.ascontiguousarray(test_embedding), self.number_of_neighbours)

            # D_test_knn, _ = self.index.search(np.ascontiguousarray(test_embedding), self.number_of_neighbours)
        if self.first_global_normalization:
            test_embedding = self.whiten.apply_py(test_embedding)
            if np.isnan(test_embedding).any():
            #     print('test nan')
            #     # breakpoint()
            #
            #     # def time_this(func, x):
            #     #     start_time = time.time()
            #     #     func(x)
            #     #     end_time = time.time()
            #     #     print(f'tot time: {end_time - start_time}')
            #
            #
            #
                test_embedding = np.nan_to_num(test_embedding)
        D_test, I_test = self.index.search(np.ascontiguousarray(test_embedding), self.number_of_neighbours)
        # test_embedding_quantized = [np.expand_dims(test_embedding[:, start_ind: start_ind + self.partial_size], 1) for start_ind in range(0, test_embedding.shape[-1], self.partial_size)]
        # test_embedding_quantized = np.concatenate(test_embedding_quantized, axis=1)
        def quantize_vec_padd_last_in_embed_list(embed, partial_size):
            # breakpoint()
            # if partial_size == embed.shape[1]:
            if partial_size == embed.shape[1] and len(embed.shape) == 2:
                return np.expand_dims(embed, 1)
            if len(embed.shape) == 2:
                embed = [np.expand_dims(embed[:, start_ind: start_ind + partial_size], 1)
                         for start_ind in range(0, embed.shape[-1], partial_size)]
            else:
                embed = [np.expand_dims(embed[:, :, start_ind: start_ind + partial_size], 2)
                         for start_ind in range(0, embed.shape[-1], partial_size)]
            # add zeros if the last vector is not in the same size as the others
            pedded_vec = np.zeros_like(embed[0])
            if len(embed[0].shape) == 3:
                pedded_vec[:, :, :embed[-1].shape[-1]] = embed[-1]
                embed[-1] = pedded_vec
                embed = np.concatenate(embed, axis=1)
            else:
                pedded_vec[:, :, :, :embed[-1].shape[-1]] = embed[-1]
                embed[-1] = pedded_vec
                embed = np.concatenate(embed, axis=2)
            return embed

        test_embedding_quantized = quantize_vec_padd_last_in_embed_list(embed=test_embedding, partial_size=self.partial_size)

        # KNN
        if do_knn or self.do_knnn_plus_knn:
            # todo need to calcualte befor normaliztion
            # self.cpu_index_knn = faiss.IndexFlatL2(self.dim)
            # self.index = faiss.index_cpu_to_all_gpus(index=self.cpu_index_knn)
            # self.whiten.reverse_transform(self.embedding) # todo get origianl
            # self.index.add(np.ascontiguousarray(self.embedding))

            # breakpoint()
            return_dict['knn'] = np.expand_dims(D_test_knn.mean(1), 1)

        # Local
        if do_local:
            neighbours_embeddings = self.embedding[I_test, :]
            # neighbours_embeddings_quantized = [np.expand_dims(neighbours_embeddings[:, :, start_ind: start_ind+self.partial_size],2) for start_ind in range(0, neighbours_embeddings.shape[-1], self.partial_size)]
            # neighbours_embeddings_quantized = np.concatenate(neighbours_embeddings_quantized, axis=2)
            neighbours_embeddings_quantized = quantize_vec_padd_last_in_embed_list(embed=neighbours_embeddings, partial_size=self.partial_size)
            test_anomlay_score_local = []
            for cur_test_ind, (cur_neighbours_embeddings, cur_test_embedding) in tqdm(enumerate(zip(neighbours_embeddings_quantized, test_embedding_quantized)), total=test_embedding_quantized.shape[0]):
                curr_score_quants = []
                for quant_ind in range(cur_test_embedding.shape[0]):
                    cur_test_embedding_quant = cur_test_embedding[quant_ind, :]
                    cur_neighbours_embeddings_quant = cur_neighbours_embeddings[:, quant_ind, :]

                    cur_neighbours_embeddings_quant = np.ascontiguousarray(cur_neighbours_embeddings_quant)
                    cur_test_embedding_quant = np.ascontiguousarray(cur_test_embedding_quant)
                    curr_whitening = faiss.PCAMatrix(d_in=self.partial_size, d_out=self.partial_size, eigen_power=self.eigen_power) # after transformation the components are multiplied by eigenvalues^eigen_power =0: no whitening =-0.5: full whitening
                    curr_whitening.epsilon = 1e-5
                    curr_whitening.train(np.ascontiguousarray(cur_neighbours_embeddings_quant))
                    # curr_eig_vals = faiss.vector_to_array(curr_whitening.eigenvalues)

                    cur_neighbours_embeddings_quant_normalized = curr_whitening.apply_py(cur_neighbours_embeddings_quant)  # TODO add transform to test as well
                    cur_test_embedding_quant_normalized = curr_whitening.apply_py(np.expand_dims(cur_test_embedding_quant, axis=0))  # TODO add transform to test as well

                    if np.isnan(cur_neighbours_embeddings_quant_normalized.min()):
                        cur_neighbours_embeddings_quant_normalized = np.nan_to_num(cur_neighbours_embeddings_quant_normalized)
                    if np.isnan(cur_test_embedding_quant_normalized.min()):
                        cur_test_embedding_quant_normalized = np.nan_to_num(cur_test_embedding_quant_normalized)
                    curr_score = defualt_dist_func_one_to_many(cur_test_embedding_quant_normalized, cur_neighbours_embeddings_quant_normalized)
                    curr_score_quants += [np.expand_dims(curr_score, axis=0)]
                curr_score_quants = np.concatenate(curr_score_quants, axis=0)
                test_anomlay_score_local += [np.expand_dims(np.array(curr_score_quants).mean(0), axis=0)]
            test_anomlay_score_local = np.concatenate(test_anomlay_score_local, axis=0)
            if self.do_knnn_plus_knn:
                test_anomlay_score_local += 0.5 * return_dict['knn'][cur_test_ind]
                # test_anomlay_score_local *= return_dict['knn'][cur_test_ind]
            return_dict['knn_local'] = self.post_dist_func(test_anomlay_score_local)

        # KNNN
        if do_knnn:
            test_anomaly_score_knnn = []
            curr_whitening = faiss.PCAMatrix(d_in=self.partial_size, d_out=self.partial_size, eigen_power=self.eigen_power)  # todo move out
            curr_whitening.epsilon = 1e-4
            for cur_test_ind, (cur_neighbours_index, cur_test_embedding) in tqdm(enumerate(zip(I_test, test_embedding_quantized)), total=test_embedding_quantized.shape[0]):
                neighbours_embeddings_of_neighbours_index = self.top_k_indexes_train[cur_neighbours_index, :]
                neighbours_embeddings_of_neighbours_embeddings = self.embedding[neighbours_embeddings_of_neighbours_index, :]
                neighbours_embeddings_of_neighbours_embeddings_quants = quantize_vec_padd_last_in_embed_list(embed=neighbours_embeddings_of_neighbours_embeddings, partial_size=self.partial_size)
                neighbours_results = []
                for neighbours_of_neighbour in neighbours_embeddings_of_neighbours_embeddings_quants:
                    sub_results = []
                    for quant_ind in range(neighbours_of_neighbour.shape[1]):
                        curr_quant_embeddings = np.ascontiguousarray(neighbours_of_neighbour[:, quant_ind, :])
                        test_quant_embeddings = np.ascontiguousarray(np.expand_dims(cur_test_embedding[quant_ind, :], axis=0))
                        if self.do_weighted_eig:
                            # breakpoint()
                            # weights = distance.cdist(test_quant_embeddings, curr_quant_embeddings)
                            weights = distance.cdist(np.expand_dims(curr_quant_embeddings[0], 0), curr_quant_embeddings)
                            eps = 1 / (curr_quant_embeddings.shape[0] / 2)
                            # weights = np.ones_like(weights)
                            weights = 1 / (weights + eps)
                            weights = weights / weights.sum()
                            W = np.zeros((weights.shape[1], weights.shape[1]))
                            np.fill_diagonal(W, weights)
                            weights = np.tile(weights.T, (1, curr_quant_embeddings.shape[1]))
                            # breakpoint()
                            weighted_curr_quant_embeddings = weights * curr_quant_embeddings
                            weighted_curr_quant_embeddings_means = np.tile(np.expand_dims(weighted_curr_quant_embeddings.sum(0), 0), (curr_quant_embeddings.shape[0],1))
                            weighted_centered_curr_quant_embeddings = curr_quant_embeddings - weighted_curr_quant_embeddings_means
                            X = weighted_centered_curr_quant_embeddings
                            # breakpoint()
                            cov_mat = X.T @ W @ X
                            eig_vals, eig_vecs = np.linalg.eig(cov_mat)  # eig_vecs[:, i] is the eigen vector of eig_vals[i]
                            eig_vals_for_norm = np.float_power(eig_vals, self.eigen_power)
                            eig_vals_for_norm = np.tile(np.expand_dims(eig_vals_for_norm, 0), (eig_vecs.shape[0], 1))
                            eign_vec_normed = eig_vecs * eig_vals_for_norm
                            curr_quant_embedding = (curr_quant_embeddings - weighted_curr_quant_embeddings_means) @ eign_vec_normed
                            test_quant_embedding = (test_quant_embeddings - weighted_curr_quant_embeddings_means) @ eign_vec_normed
                            # curr_quant_embedding = (curr_quant_embeddings ) @ eign_vec_normed # TODO check which is better
                            # test_quant_embedding = (test_quant_embeddings )@ eign_vec_normed

                        else:
                            curr_whitening.train(curr_quant_embeddings)
                            curr_quant_embedding = curr_whitening.apply_py(np.expand_dims(curr_quant_embeddings[0], axis=0))
                            test_quant_embedding = curr_whitening.apply_py(test_quant_embeddings)
                        if np.isnan(curr_quant_embedding).any() or np.isnan(test_quant_embedding).any():
                            # print('nan')
                            curr_quant_embedding = np.nan_to_num(curr_quant_embedding)
                            test_quant_embedding = np.nan_to_num(test_quant_embedding)
                        # curr_score = defualt_dist_func(test_quant_embedding, curr_quant_embedding)
                        curr_score = defualt_dist_func(test_quant_embedding, np.expand_dims(curr_quant_embedding[0],0))  # todo tmp !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                        # curr_score = defualt_dist_func(test_quant_embedding, np.expand_dims(weighted_curr_quant_embeddings_means[0],0))  # todo tmp !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        sub_results += [curr_score]
                    neighbours_results += [np.expand_dims(np.array(sub_results).mean(), axis=0)]
                # breakpoint()
                if self.do_knnn_plus_knn:
                    neighbours_results += 0.5 * return_dict['knn'][cur_test_ind]
                    # neighbours_results *= return_dict['knn'][cur_test_ind]
                test_anomaly_score_knnn += [np.expand_dims(np.concatenate(neighbours_results, axis=0), axis=0)]

            # breakpoint()
            test_anomaly_score_knnn = np.concatenate(test_anomaly_score_knnn, axis=0)
            return_dict['knnn'] = self.post_dist_func(test_anomaly_score_knnn)

        # Global
        if do_global:
            # test_anomaly_score_knn_global = [float(anomaly_dists.mean()) for anomaly_dists in D_test]
            test_anomaly_score_knn_global = D_test
            return_dict['knn_global'] = self.post_dist_func(test_anomaly_score_knn_global)





        # # reorder the embedding if do_reorder is set to True
        # self.test_embedding = test_embedding[:, self.new_order] if self.do_reorder else test_embedding
        # return_dict = {}
        # self.dist_mat_test = calc_dist_matrix(x=self.test_embedding, y=self.embedding)
        # self.top_k_indexes_test_to_train = get_k_neighbours_indexes_for_each_sample(dist_mat=self.dist_mat_test,
        #                                                                             top_k=self.number_of_neighbours)
        # self.top_k_indexes_test_to_train = get_k_neighbours_indexes_for_each_sample_faiss(embedding_train=self.embedding,
        #                                                                                   embedding_test=self.test_embedding,
        #                                                                             top_k=self.number_of_neighbours)
        # if do_knnn:
        #     self.dist_results = normalized_dist_by_partial_eigs(embedding_test=self.test_embedding,
        #                                                         embedding_train=self.embedding,
        #                                                         neighbours_indices_mat=self.top_k_indexes_test_to_train,
        #                                                         neighbours_of_neighbours_indices_mat=self.top_k_indexes_train,
        #                                                         partial_size=self.partial_size,
        #                                                         partial_eig_vec_s=self.partial_vec_s,
        #                                                         partial_eig_val_s=self.partial_val_s)
        #     return_dict['knnn'] = self.post_dist_func(self.dist_results)
        # calculate knn local normalization
        # if do_local:
        #     self.dist_results_local = normalized_dist_by_partial_eigs(embedding_test=self.test_embedding,
        #                                                               embedding_train=self.embedding,
        #                                                               neighbours_indices_mat=self.top_k_indexes_test_to_train,
        #                                                               neighbours_of_neighbours_indices_mat=self.top_k_indexes_train,
        #                                                               partial_size=self.partial_size,
        #                                                               do_local=True)
        #
        #     return_dict['knn_local'] = self.post_dist_func(self.dist_results_local)

        # if do_global:
        #     # calculate results based on global normalization
        #     partial_vec_global_s = np.tile(np.expand_dims(self.partial_vec_global_s, 0),
        #                                    (self.embedding.shape[0], 1, 1, 1))
        #     partial_val_global_s = np.tile(np.expand_dims(self.partial_val_global_s, 0),
        #                                    (self.embedding.shape[0], 1, 1))
        #     self.dist_results_global = normalized_dist_by_partial_eigs(embedding_test=self.test_embedding,
        #                                                                embedding_train=self.embedding,
        #                                                                neighbours_indices_mat=self.top_k_indexes_test_to_train,
        #                                                                neighbours_of_neighbours_indices_mat=self.top_k_indexes_train,
        #                                                                partial_eig_vec_s=partial_vec_global_s,
        #                                                                partial_eig_val_s=partial_val_global_s)
        #     return_dict['knn_global'] = self.post_dist_func(self.dist_results_global)
        #
        # # calculate results based on regular knn
        # if do_knn:
        #     knn_dists = calc_dist_matrix(self.test_embedding, self.embedding)
        #     knn_top_k_closest_dist = np.partition(a=knn_dists, kth=self.number_of_neighbours, axis=1)[:, :self.number_of_neighbours]
        #     self.dist_results_knn = knn_top_k_closest_dist.mean(1)
        #     return_dict['knn'] = self.post_dist_func(self.dist_results_knn)

        return return_dict


# TODO add main example of knnn run
